#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#ifndef SJIS_UNICODE_H
#include "sjis_unicode.h"
#endif


unsigned long CalcTableChecksum(unsigned long* table, unsigned long length);
long IntLog2(long);



struct char_map {
	unsigned short		map;
	bool				valid;
};

struct cmap_record {
	unsigned short		platform;
	unsigned short		encoding;
	unsigned long		table_offset;
	unsigned long		table_length;
};




main(int argc, char* argv[])
{
	if (argc != 2) {
		printf("Usage: ttc2ttf filename\n");
		exit(2);
	}

	sjis_unicode_setup();

	char*		file_name = argv[1];
	FILE*		in_file = fopen(file_name, "rt");
	if (in_file == NULL) {
		printf("%s not found", file_name);
		exit(2);
	}


	struct	stat	info;
	fstat(fileno(in_file), &info);
	size_t			buffer_size = info.st_size;
	char*			buffer = (char*) malloc(buffer_size);
	fread(buffer, sizeof(char), buffer_size, in_file);
	fclose(in_file);



	//	find cmap
	unsigned short	table_count = *(unsigned short *) &buffer[4];
	unsigned long	cmap_offset, cmap_length = 0, cmap_header_offset;
	for (unsigned short i = 0; i < table_count; i++)
		if (*(unsigned long*) &buffer[0x0C + i * 0x10] == 'cmap') {
			cmap_offset = *(unsigned long*) &buffer[0x0C + 0x08 + i * 0x10];
			cmap_length = *(unsigned long*) &buffer[0x0C + 0x0C + i * 0x10];
			cmap_header_offset = 0x0C + i * 0x10;
			break;
		}
	if (cmap_length == 0) {
		printf("cmap not found\n");
		exit(2);
	}


	//	read cmap
	unsigned short	cmap_count = *(unsigned short*) &buffer[cmap_offset + 0x02];
	cmap_record*	map_array = (cmap_record*) malloc(cmap_count * sizeof(cmap_record));
	unsigned long	sjis_offset = 0;
	for (long i = 0; i < cmap_count; i++) {
		map_array[i].platform = *(unsigned short*) &buffer[cmap_offset + 0x04 + i * 0x08];
		map_array[i].encoding = *(unsigned short*) &buffer[cmap_offset + 0x04 + i * 0x08 + 0x02];
		map_array[i].table_offset = *(unsigned long*) &buffer[cmap_offset + 0x04 + i * 0x08 + 0x04];
		map_array[i].table_length = *(unsigned short*) &buffer[cmap_offset + map_array[i].table_offset + 0x02];
		if (map_array[i].platform == 3 && map_array[i].encoding == 2)
			sjis_offset = cmap_offset + map_array[i].table_offset;
	}
	if (sjis_offset == 0) {
		printf("sjis map not found\n");
		exit(2);
	}
	if (*(unsigned short*) &buffer[sjis_offset] != 4) {
		printf("bad sjis cmap\n");
		exit(2);
	}



	//	make unicode map
	char_map*		unicode_map = (char_map*) malloc(0x10000 * sizeof(char_map));
	for (long i = 0; i < 0x10000; i++)
		unicode_map[i].valid = false;

	unsigned short		cmap_seg_count = *(unsigned short*) &buffer[sjis_offset + 0x06] / 2;
	unsigned short*		cmap_end_array = (unsigned short*) &buffer[sjis_offset + 0x0E];
	unsigned short*		cmap_start_array = cmap_end_array + cmap_seg_count + 1;
	unsigned short*		cmap_delta_array = cmap_start_array + cmap_seg_count;
	unsigned short*		cmap_offset_array = cmap_delta_array + cmap_seg_count;
	unsigned short*		cmap_glyphld_array = cmap_offset_array + cmap_seg_count;
	*(cmap_end_array + cmap_seg_count) = 0;

	for (long i = 0; i < cmap_seg_count; i++)
		if (cmap_end_array[i] != 0xFFFF) {
			if (cmap_offset_array[i] == 0) {
				for (unsigned long j = cmap_start_array[i]; j <= cmap_end_array[i]; j++) {
					char_map*	map = &unicode_map[sjis_to_unicode(j)];
					map->map = j + cmap_delta_array[i];
					map->valid = true;
				}
			} else {
				for (unsigned long j = cmap_start_array[i]; j <= cmap_end_array[i]; j++) {
					char_map*	map = &unicode_map[sjis_to_unicode(j)];
					map->map = cmap_delta_array[i] + cmap_offset_array[i + cmap_offset_array[i] / 2 + j - cmap_start_array[i]];
					map->valid = true;
				}
			}
		}


	unsigned short	seg_count = 0;
	long			index = 0;
	while (index < 0x10000) {
		if (unicode_map[index].valid) {
			long	next_index = index + 1;
			unsigned short	last = unicode_map[index].map;
			while (next_index < 0x10000 && unicode_map[next_index].valid && unicode_map[next_index].map == last + 1) {
				last++;
				next_index++;
			}
			index = next_index;
			seg_count++;
		} else
			index++;
	}
	seg_count++;


	unsigned short	entry_selector = IntLog2(seg_count);
	unsigned short	search_range = 2 * pow(2, entry_selector);
	unsigned long	new_cmap_length = 0x04 + 0x08 + 16 + 4 * seg_count * 2;
	for (long i = 0; i < cmap_count; i++)
		if (map_array[i].platform != 3 || map_array[i].encoding != 2)
			new_cmap_length += 0x08 + map_array[i].table_length;
	unsigned long	uni_cmap_length = 16 + 4 * seg_count * 2;
	size_t			new_buffer_size = buffer_size + ((new_cmap_length + 3) & ~3) - ((cmap_length + 3) & ~3);
	char*			new_buffer = (char*) malloc(new_buffer_size);
	memcpy(new_buffer, buffer, cmap_offset);
	memcpy(&new_buffer[cmap_offset + ((new_cmap_length + 3) & ~3)], &buffer[cmap_offset + ((cmap_length + 3) & ~3)], buffer_size - cmap_offset - ((cmap_length + 3) & ~3));


	unsigned short*		sptr = (unsigned short*) &new_buffer[cmap_offset];
	*sptr++ = 0;
	*sptr++ = cmap_count;

	unsigned long		current_offset = 0x04 + cmap_count * 0x08;
	for (long i = 0; i < cmap_count; i++)
		if (map_array[i].platform != 3 || map_array[i].encoding != 2) {
			*sptr++ = map_array[i].platform;
			*sptr++ = map_array[i].encoding;
			*(unsigned long *) sptr = current_offset;
			sptr += 2;
			current_offset += map_array[i].table_length;
		}
	*sptr++ = 3;
	*sptr++ = 1;
	*(unsigned long *) sptr = current_offset;
	sptr += 2;
	char*				cptr = new_buffer + cmap_offset + 0x04 + cmap_count * 0x08;
	for (long i = 0; i < cmap_count; i++)
		if (map_array[i].platform != 3 || map_array[i].encoding != 2) {
			memcpy(cptr, buffer + cmap_offset + map_array[i].table_offset, map_array[i].table_length);
			cptr += map_array[i].table_length;
		}
	sptr = (unsigned short*) cptr;
	*sptr++ = 4;
	*sptr++ = uni_cmap_length;
	*sptr++ = 0;
	*sptr++ = 2 * seg_count;
	*sptr++ = search_range;
	*sptr++ = entry_selector;
	*sptr++ = (2 * seg_count) - search_range;

	cmap_end_array = sptr;
	cmap_start_array = cmap_end_array + seg_count + 1;
	cmap_delta_array = cmap_start_array + seg_count;
	cmap_offset_array = cmap_delta_array + seg_count;

	index = 0;
	while (index < 0x10000) {
		if (unicode_map[index].valid) {
			long	next_index = index + 1;
			unsigned short	last = unicode_map[index].map;
			while (next_index < 0x10000 && unicode_map[next_index].valid && unicode_map[next_index].map == last + 1) {
				last++;
				next_index++;
			}

			*cmap_end_array++ = next_index - 1;
			*cmap_start_array++ = index;
			*cmap_delta_array++ = unicode_map[index].map - (unsigned short) index;
			*cmap_offset_array++ = 0;

			index = next_index;
		} else
			index++;
	}
	*cmap_end_array = 0xFFFF;
	*cmap_start_array = 0xFFFF;
	*cmap_delta_array = 1;
	*cmap_offset_array = 0;


	unsigned long*	lptr = (unsigned long*) (new_buffer + cmap_header_offset);
	*lptr++ = 'cmap';
	*lptr++ = CalcTableChecksum((unsigned long*) &new_buffer[cmap_offset], new_cmap_length);
	*lptr++ = cmap_offset;
	*lptr++ = new_cmap_length;

	for (long i = 0x0C; i < (0xC + 0x10 * table_count); i += 0x10)
		if (*(unsigned long*) &new_buffer[i + 0x8] > cmap_offset)
			*(unsigned long*) &new_buffer[i + 0x8] += ((new_cmap_length + 3) & ~3) - ((cmap_length + 3) & ~3);


	FILE*	out_file = fopen(file_name, "wt");
	fwrite(new_buffer, sizeof(char), new_buffer_size, out_file);
	fclose(out_file);

	free(buffer);
	free(new_buffer);
	free(unicode_map);
	free(map_array);
}


unsigned long
CalcTableChecksum(unsigned long* table, unsigned long length)
{
	unsigned long	sum = 0L;
	unsigned long*	end_ptr = table + ((length + 3) & ~3) / sizeof(unsigned long);
	while (table < end_ptr)
		sum += *table++;
	return sum;
}

long IntLog2(long value)
{
	long	cur = 0;
	while (!(pow(2, cur) <= value && pow(2, cur + 1) > value)) cur++;
	return cur;
}

